use clap::{Parser, Subcommand};
use std::{fs, path::PathBuf};

use crate::config::Config;
use crate::error::Error;
use crate::migrator::Migrator;
use crate::setup::Initializer;

pub async fn run_cli() -> Result<(), Error> {
    let cli = Cli::parse();

    let migration_dir = cli
        .migration_dir
        .expect("Migration directory not specified");

    match cli.subcommand {
        MigrationCommand::Generate { migration_name } => {
            Migrator::generate(migration_dir, &migration_name)?;
            return Ok(());
        }
        _ => {}
    };

    let yml = fs::File::open(cli.config.expect("Path to 'default.yml' not specified"))
        .expect("Failed to open 'default.yml'");
    let config: Config = serde_yaml::from_reader(yml).expect("Failed to parse yaml");
    let scylla_conf = config
        .scylla
        .expect("ScyllaDB config not found in 'default.yml'");

    match cli.subcommand {
        MigrationCommand::Up { num } => {
            Migrator::new(migration_dir, &scylla_conf)
                .await?
                .up(num)
                .await?
        }
        MigrationCommand::Down { num } => {
            Migrator::new(migration_dir, &scylla_conf)
                .await?
                .down(num)
                .await?
        }
        MigrationCommand::Setup {
            threads,
            note_since,
            note_skip,
            no_confirm,
        } => {
            let initializer =
                Initializer::new(&scylla_conf, &config.db, note_since, note_skip, no_confirm)
                    .await?;
            initializer.setup(threads).await?;
        }
        _ => {}
    };

    Ok(())
}

#[derive(Parser)]
#[clap(version, about = "ScyllaDB Migration Tool")]
pub(crate) struct Cli {
    #[clap(
        value_parser,
        global = true,
        short = 'c',
        long,
        help = "Path to 'default.yml'"
    )]
    pub config: Option<PathBuf>,

    #[clap(
        value_parser,
        global = true,
        short = 'd',
        long,
        help = "Directory to store migration files"
    )]
    pub migration_dir: Option<PathBuf>,

    #[clap(subcommand)]
    pub subcommand: MigrationCommand,
}

#[derive(Subcommand)]
pub(crate) enum MigrationCommand {
    #[clap(about = "Generate a new, empty migration", display_order = 10)]
    Generate {
        #[clap(
            value_parser,
            required = true,
            help = "Name of the new migration",
            display_order = 11
        )]
        migration_name: String,
    },
    #[clap(about = "Apply pending migrations", display_order = 20)]
    Up {
        #[clap(
            value_parser,
            short,
            long,
            help = "Number of pending migrations to apply"
        )]
        num: Option<u32>,
    },
    #[clap(
        value_parser,
        about = "Rollback applied migrations",
        display_order = 30
    )]
    Down {
        #[clap(
            value_parser,
            short,
            long,
            default_value = "1",
            help = "Number of applied migrations to be rolled back",
            display_order = 31
        )]
        num: u32,
    },
    #[clap(about = "Set up PostgreSQL and ScyllaDB", display_order = 40)]
    Setup {
        #[clap(
            value_parser,
            short,
            long,
            default_value = "1",
            help = "Number of threads used to copy",
            display_order = 41
        )]
        threads: u32,
        #[clap(
            value_parser,
            long,
            help = "Note ID to begin copying",
            display_order = 42
        )]
        note_since: Option<String>,
        #[clap(
            value_parser,
            long,
            default_value = "0",
            help = "The number of notes to be skipped while copying",
            display_order = 43
        )]
        note_skip: u64,
        #[clap(
            value_parser,
            long,
            default_value = "false",
            help = "Does not confirm before the process begins",
            display_order = 44
        )]
        no_confirm: bool,
    },
}
